unit PanelWiz;

{
  Insert a panel as the parent of the selected components. The size of
  the new panel is set to fit around the selected components. If the selection
  is a single control with an Align property, set the Align property to
  alClient, so it always fills the new panel.

  Copyright  1998 Tempest Software, Inc.
}

interface

uses Windows, SysUtils, Classes, Graphics, Controls, Dialogs, Forms,
  ExptIntf, ToolIntf, EditIntf, Etk;

type
  TInsertPanelWizard = class(TEtkModule)
    EditInsertPanel: TEtkMenuItem;
    procedure EditInsertPanelClick(Sender: TObject);
  private
    procedure ChangeParent(Control, Parent: TIComponentInterface);
    function GetBounds(Control: TIComponentInterface; var Rect: TRect): Boolean;
    function GetInteger(Component: TIComponentInterface; const PropName: string; var Value: Integer): Boolean;
    function GetParent(Form: TIFormInterface; Control: TIComponentInterface): TIComponentInterface;
    procedure InsertParent(Form: TIFormInterface);
    procedure Reposition(Control, Parent: TIComponentInterface);
    function SelectionIsForm(Form: TIFormInterface; Selection: TIComponentInterface): Boolean;
    procedure SetAlign(Selection: TIComponentInterface);
  end;

var
  InsertPanelWizard: TInsertPanelWizard;

procedure Register;

implementation

{$R *.DFM}

resourcestring
  sPleaseSelect = 'No form! You must edit a form to use the Insert Panel Wizard';
  sNoSelection = 'No selection! Please select one or more components';
  sTPanel = 'TPanel'; { so a smart user can change the control type }

{ Change the Parent property of a control. Because the Parent property is
  not published, a DLL expert cannot do this using the Open Tools API.
  Instead, cast the control's handle to TControl, which requires that
  this expert reside in a package. }
procedure TInsertPanelWizard.ChangeParent(Control, Parent: TIComponentInterface);
var
  ParentHandle: TObject;
begin
  ParentHandle := Parent.GetComponentHandle;
  if Control.IsTControl and (ParentHandle is TWinControl) then
    TControl(Control.GetComponentHandle).Parent := TWinControl(ParentHandle);
end;

{ Get the bounds of a control and store them in Rect. Return True if
  successful, False if the control does not publish Left, Top, Height,
  and Width properties. Also make sure no one created a perverse control
  that has non-integer types for these properties. }
function TInsertPanelWizard.GetBounds(Control: TIComponentInterface; var Rect: TRect): Boolean;
var
  Width, Height: Integer;
begin
  Result := False;

  if not GetInteger(Control, 'Left', Rect.Left) then
    Exit;
  if not GetInteger(Control, 'Top', Rect.Top) then
    Exit;
  if not GetInteger(Control, 'Width', Width) then
    Exit;
  if not GetInteger(Control, 'Height', Height) then
    Exit;

  Rect.Right := Rect.Left + Width;
  Rect.Bottom := Rect.Top + Height;
  Result := True;
end;

{ Get an integer property. Return True for success, False if the property
  is not published or does not have an integer type. Store the integer
  property value in the Value variable, but only if the function returns True. }
function TInsertPanelWizard.GetInteger(Component: TIComponentInterface; const PropName: string; var Value: Integer): Boolean;
begin
  if Component.GetPropTypeByName(PropName) <> ptInteger then
    Result := False
  else
    Result := Component.GetPropValueByName(PropName, Value);
end;

{ TIComponentInterface.GetParent always returns nil. The GetParent function
  laboriously searches the form for the real parent. The caller must free the
  parent interface. }
function TInsertPanelWizard.GetParent(Form: TIFormInterface; Control: TIComponentInterface): TIComponentInterface;

  { Recursively search the children of Parent for a child whose handle
    is that of Control. That means Parent is the parent of Control, so
    return Parent. Return nil if the parent is not found. }
  function FindRealParent(Parent: TIComponentInterface): TIComponentInterface;
  var
    Child: TIComponentInterface;
    I: Integer;
  begin
    Result := nil;
    for I := 0 to Parent.GetControlCount-1 do
    begin
      Child := Parent.GetControl(I);
      try
        if Child.GetComponentHandle = Control.GetComponentHandle then
        begin
          Result := Parent;
          Result.AddRef;
        end
        else
          Result := FindRealParent(Child);
        if Result <> nil then
          Exit;
      finally
        Child.Free;
      end;
    end;
  end;

var
  FormComponent: TIComponentInterface;
begin
  FormComponent := Form.GetFormComponent;
  try
    Result := FindRealParent(FormComponent);
  finally
    FormComponent.Free;
  end;
end;

{ Insert a TPanel as the parent control of the current selection. Start by
  determining the bounding box for the selected controls, and make the new
  panel that size. Then change the parent of each selected control to be
  the panel. If the controls all have a common parent, make the panel a
  child of that parent. Otherwise, the new panel is a child of the form.

  Note that if the controls do not have a common parent, the bounding box
  will be nonsensical because the Left and Top properties are relative
  to a control's parent. This function does not try to normalize the positions
  with respect to the form. This means that controls will appear to move
  on the form. If you don't like it, you can fix it.
}

procedure TInsertPanelWizard.InsertParent(Form: TIFormInterface);
var
  Bounds: TRect;
  Rect: TRect;
  NewPanel: TIComponentInterface;
  Control: TIComponentInterface;
  Parent: TIComponentInterface;
  CommonParent: TIComponentInterface;
  HaveCommonParent: Boolean;
  Selections: TList;
  I: Integer;
begin
  { Get the bounding box of the current selection. }
  Bounds.Left := 0;
  Bounds.Top := 0;
  Bounds.Right := 0;
  Bounds.Bottom := 0;
  HaveCommonParent := True;

  CommonParent := nil;
  try
    Selections := TList.Create;
    try
      for I := 0 to Form.GetSelCount-1 do
      begin
        Control := Form.GetSelComponent(I);
        try
          { Save a list of the selected controls because creating a new control
            changes the selection list. }
          Selections.Add(Control.GetComponentHandle);

          { Getting the bounding box is easy. }
          if GetBounds(Control, Rect) then
            UnionRect(Bounds, Bounds, Rect);

          { Now see if this control's parent is the same as the other controls' parents. }
          if HaveCommonParent then
          begin
            Parent := GetParent(Form, Control);
            try
              if CommonParent = nil then
              begin
                { This is the first control, so save its parent as the common parent.}
                CommonParent := Parent;
                if CommonParent <> nil then
                  CommonParent.AddRef;
              end
              else if (Parent <> nil) and (CommonParent.GetComponentHandle <> Parent.GetComponentHandle) then
              begin
                { Another control has a different parent, so give up trying to
                  keep track of the common parent. }
                CommonParent.Free;
                CommonParent := nil;
                HaveCommonParent := False;
              end;
            finally
              Parent.Free;
            end;
          end;
        finally
          Control.Free;
        end;
      end;

      { Create the new panel. If CommonParent is nil, the new panel will be
        a child of the form. }
      with Bounds do
        NewPanel := Form.CreateComponent(CommonParent, sTPanel, Left, Top, Right-Left, Bottom-Top);

      { Reparent the selection to be children of the new panel. Change the
        controls' positions to be relative to the new panel. }
      try
        for I := 0 to Selections.Count-1 do
        begin
          Control := Form.GetComponentFromHandle(Selections[I]);
          try
            ChangeParent(Control, NewPanel);
            Reposition(Control, NewPanel);
          finally
            Control.Free;
          end;
        end;
      finally
        NewPanel.Free;
      end;
    finally
      Selections.Free;
    end;
  finally
    CommonParent.Free;
  end;
end;

{ Change the Left and Top properties of Control so they are relative to
  the positon of Parent. }
procedure TInsertPanelWizard.Reposition(Control, Parent: TIComponentInterface);
var
  ControlLeft, ControlTop: Integer;
  ParentLeft, ParentTop: Integer;
begin
  if not GetInteger(Control, 'Left', ControlLeft) then
    Exit;
  if not GetInteger(Control, 'Top', ControlTop) then
    Exit;
  if not GetInteger(Parent, 'Left', ParentLeft) then
    Exit;
  if not GetInteger(Parent, 'Top', ParentTop) then
    Exit;

  ControlLeft := ControlLeft - ParentLeft;
  ControlTop  := ControlTop  - ParentTop;

  Control.SetPropByName('Left', ControlLeft);
  Control.SetPropByName('Top',  ControlTop);
end;

{ Return True if Selection is the form component, which means the user has
  not selected any real components. Obviously, you cannot insert a panel
  as the parent of the form. Return False if the selection is nil or is
  otherwise not the form component. }
function TInsertPanelWizard.SelectionIsForm(Form: TIFormInterface; Selection: TIComponentInterface): Boolean;
var
  FormComponent: TIComponentInterface;
begin
  if Selection = nil then
    Result := False
  else
  begin
    FormComponent := Form.GetFormComponent;
    try
      Result := FormComponent.GetComponentHandle = Selection.GetComponentHandle;
    finally
      FormComponent.Free;
    end;
  end;
end;

{ Set the Align property of the old selection to alClient, so it completely
  fills the new panel. If the control does not have an Align property whose
  type is an enumeration, do nothing. }
procedure TInsertPanelWizard.SetAlign(Selection: TIComponentInterface);
var
  AlClientVar: TAlign;
begin
  if Selection.GetPropTypeByName('Align') = ptEnumeration then
  begin
    { SetPropByName requires a variable for the property value. }
    AlClientVar := alClient;
    Selection.SetPropByName('Align', AlClientVar);
  end;
end;

{ The user selected the Insert Panel menu item. Create a new panel as the parent
  of the selected components. As a special case, fi the user selected a single
  panel, change the old panel's Align property to alClient. }
procedure TInsertPanelWizard.EditInsertPanelClick(Sender: TObject);
var
  Module: TIModuleInterface;
  Form: TIFormInterface;
  Selection: TIComponentInterface;
begin
  with ToolServices do
    { If the user is editing a form, GetCurrentFile return the path to
      the DFM file, but we need the path to the PAS file. If the user is
      not editing a DFM file, punt. }
    if CompareText(ExtractFileExt(GetCurrentFile), '.dfm') = 0 then
      Module := GetModuleInterface(ChangeFileExt(GetCurrentFile, '.pas'))
    else
      Module := nil;

  if Module = nil then
    raise Exception.Create(sPleaseSelect);

  try
    Form := Module.GetFormInterface;
    if Form = nil then
      raise Exception.Create(sPleaseSelect);

    try
      { See if the selection is a single component. If so, save its component
        interface, to fix its Align property later. Inserting the new panel
        changes the selection, so save the selected panel first. }
      case Form.GetSelCount of
      0:    raise Exception.Create(sNoSelection);
      1:    Selection := Form.GetSelComponent(0)
      else  Selection := nil;
      end;

      try
        if SelectionIsForm(Form, Selection) then
          raise Exception.Create(sNoSelection);

        { Create and insert the panel. }
        InsertParent(Form);

        { Fix the previously selected component's Align property. }
        if Selection <> nil then
          SetAlign(Selection);
      finally
        Selection.Free;
      end;
    finally
      Form.Free;
    end;
  finally
    Module.Free;
  end;
end;

procedure Register;
begin
  TInsertPanelWizard.CreateAndRegister(InsertPanelWizard);
end;

end.
